/*
 * Copyright (c) 2007      Gustavo Pichorim Boiko <gustavo.boiko@kdemail.net>
 * Copyright (c) 2002,2003 Hamish Rodda <rodda@kde.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <kglobalaccel.h>
#include <kapplication.h>
#include <kpopupmenu.h>
#include <khelpmenu.h>
#include <kiconloader.h>
#include <kaction.h>
#include <kstdguiitem.h>
#include <qtooltip.h>
#include <kcmultidialog.h>
#include "krandrtray.h"
#include "krandrpassivepopup.h"
#include "krandrtray.moc"
#include "legacyrandrscreen.h"
#include "randrscreen.h"
#include "randroutput.h"
#include "randrmode.h"

#include <kdebug.h>
KRandRSystemTray::KRandRSystemTray(RandRDisplay *dpy, QWidget* parent, const char *name)
: KSystemTray(parent, name)
, m_popupUp(false), m_display(dpy)
, m_help(new KHelpMenu(parent, KGlobal::instance()->aboutData(), false, actionCollection()))
{
	setPixmap(KSystemTray::loadIcon("randr"));
	setAlignment(Qt::AlignHCenter | Qt::AlignVCenter);
	connect(this, SIGNAL(quitSelected()), kapp, SLOT(quit()));
	QToolTip::add(this, i18n("Screen resize & rotate"));

	KConfig cfg("kcmrandrrc");
	cfg.setGroup("General");

	m_runStartup = cfg.readBoolEntry("AutoStart", true);
	KGlobalAccel *accel = new KGlobalAccel(this);
	accel->insert( "DisplaySwitch", i18n("Switch displays"),
		   i18n("Switch between active displays."),
		        CTRL+SHIFT+Key_S, KKey::QtWIN+CTRL+Key_S, this, SLOT(slotSwitchDisplay()) );

	accel->readSettings(&cfg);
	accel->updateConnections();
}

void KRandRSystemTray::mousePressEvent(QMouseEvent* e)
{
	// Popup the context menu with left-click
	if (e->button() == LeftButton) {
		contextMenuAboutToShow(contextMenu());
		contextMenu()->popup(e->globalPos());
		e->accept();
		return;
	}

	KSystemTray::mousePressEvent(e);
}

void KRandRSystemTray::contextMenuAboutToShow(KPopupMenu* menu)
{
	int lastIndex = 0;

	menu->clear();
	menu->setCheckable(true);

	if (!m_display->isValid()) {
		lastIndex = menu->insertItem(i18n("Required X Extension Not Available"));
		menu->setItemEnabled(lastIndex, false);

	} else {
		m_screenPopups.clear();
		for (int s = 0; s < m_display->numScreens(); s++) {
			m_display->setCurrentScreen(s);
			if (s == m_display->screenIndexOfWidget(this)) {
				/*lastIndex = menu->insertItem(i18n("Screen %1").arg(s+1));
				menu->setItemEnabled(lastIndex, false);*/
			} else {
#ifdef HAS_RANDR_1_2
			    if (RandR::has_1_2)
				    m_display->currentScreen()->loadSettings();
#endif
				KPopupMenu* subMenu = new KPopupMenu(menu, QString("screen%1").arg(s+1).latin1());
				m_screenPopups.append(subMenu);
				populateMenu(subMenu);
				lastIndex = menu->insertItem(i18n("Screen %1").arg(s+1), subMenu);
				connect(subMenu, SIGNAL(activated(int)), SLOT(slotScreenActivated()));
			}
		}

		m_display->setCurrentScreen(m_display->screenIndexOfWidget(this));
#ifdef HAS_RANDR_1_2
		if (RandR::has_1_2)
			m_display->currentScreen()->loadSettings();
#endif
		populateMenu(menu);
	}

	menu->insertSeparator();
	KGlobal::locale()->insertCatalogue("mandriva-kde-translation");
	lastIndex = menu->insertItem(i18n( "Start KRandR when KDE starts" ));
	menu->connectItem(lastIndex, this, SLOT(slotToggleRunStartup()));
	menu->setItemChecked( lastIndex, m_runStartup );

	KAction *actPrefs = new KAction( i18n( "Configure Display..." ),
		SmallIconSet( "configure" ), KShortcut(), this, SLOT( slotPrefs() ),
		actionCollection() );
	actPrefs->plug( menu );

	menu->insertItem(SmallIcon("help"),KStdGuiItem::help().text(), m_help->menu());
	KAction *quitAction = actionCollection()->action(KStdAction::name(KStdAction::Quit));
	quitAction->plug(menu);
}

void KRandRSystemTray::slotScreenActivated()
{
	m_display->setCurrentScreen(m_screenPopups.findIndex(const_cast<KPopupMenu*>(static_cast<const KPopupMenu*>(sender()))));
}

void KRandRSystemTray::configChanged()
{
	m_display->refresh();

	static bool first = true;

	if (!first)
	{
		QString message;
#ifdef HAS_RANDR_1_2
		if (RandR::has_1_2)
			// TODO: display config changed message
			message = "Screen config changed";
		else
#endif
			message = m_display->currentLegacyScreen()->changedMessage();

		KRandrPassivePopup::message(
		i18n("Screen configuration has changed"),
		message, SmallIcon("window_fullscreen"),
		this, "ScreenChangeNotification");
	}

	first = false;
}

void KRandRSystemTray::populateMenu(KPopupMenu* menu)
{
	int lastIndex;
#ifdef HAS_RANDR_1_2
	if (RandR::has_1_2) 
	{
		OutputMap outputs = m_display->currentScreen()->outputs();
		if (outputs.count() <= 0)
			return;

		RandRScreen *screen = m_display->currentScreen();
		Q_ASSERT(screen);

		// if the outputs are unified, do not show output-specific size
		// changing options in the tray.
		if (screen->outputsUnified() && screen->connectedCount() > 1)
		{
			SizeList sizes = screen->unifiedSizes();

			if (sizes.count())
			{
				// populate unified sizes
				QSize currentSize;
				currentSize = screen->rect().size();

				menu->insertTitle(SmallIcon("window_fullscreen"), i18n("Screen Size"));
				populateSizes(menu, sizes, currentSize, screen, SLOT(slotResizeUnified(int)));

				// if the outputs are unified, we can rotate the screen on all outputs
				int rotations = screen->unifiedRotations();
				if (rotations != RandR::Rotate0)
				{
					menu->insertTitle(SmallIcon("view-refresh"), i18n("Orientation"));
					int rotation = RandR::Rotate0;
					OutputMap::iterator it;
					for (it = outputs.begin(); it != outputs.end(); ++it)
					{
						RandROutput *output = *it;
						if (output->isActive())
						{
							 rotation = output->rotation();
							 break;
						}
					}
				}
			}
		}
		else
		{
			if (screen->connectedCount() != 1)
				menu->insertTitle(SmallIcon("view-fullscreen"), i18n("Outputs"));

			OutputMap::iterator it;
			for (it = outputs.begin(); it != outputs.end(); ++it)
			{
				RandROutput *output = *it;
				if (output->isConnected())
				{
					KPopupMenu *outputMenu;
					if (screen->connectedCount() == 1)
						outputMenu = menu;
					else
						outputMenu = new KPopupMenu(menu);

					QSize currentSize = output->rect().size();
					// if the output is rotated 90 or 270, the returned rect is inverted
					// so we need to invert the size before comparing
					if (output->rotation() & (RandR::Rotate90 | RandR::Rotate270))
						currentSize = QSize(currentSize.height(), currentSize.width());

					populateSizes(outputMenu, output->sizes(), currentSize, output, SLOT(slotChangeSize(int)));

					// if there is only one output active, do not show the disable option
					// this prevents the user from doing wrong things ;)
					if (screen->activeCount() != 1)
					{
						lastIndex = outputMenu->insertItem(i18n("Disable"));
						outputMenu->connectItem(lastIndex, output, SLOT(slotDisable()));
						if (output->crtc() == None)
							outputMenu->setItemChecked(lastIndex, true);
					}

					// Display the rotations
					int rotations = output->rotations();
					// Don't display the rotation options if there is no point (ie. none are supported)
					// XFree86 4.3 does not include rotation support.
					if (rotations != RandR::Rotate0)
					{
						outputMenu->insertTitle(SmallIcon("view-refresh"), i18n("Orientation"));
						populateRotations(outputMenu, rotations, output->rotation(), output, SLOT(slotChangeRotation(int)));
					}

					// refresh rate
					RateList rates = output->refreshRates();
					if (rates.count())
					{
						outputMenu->insertTitle(SmallIcon("chronometer"), i18n("Refresh Rate"));
						populateRates(outputMenu, rates, output->refreshRate(), output, SLOT(slotChangeRefreshRate(int)));
					}

					if (screen->connectedCount() != 1)
						menu->insertItem(SmallIcon(output->icon()), i18n(output->name()), outputMenu);
				}
			}
		}
		// if there is more than one output connected, give the option to unify the outputs
		if (screen->connectedCount() != 1)
		{
			menu->insertSeparator();
			lastIndex = menu->insertItem( i18n("Unify Outputs"), screen, SLOT(slotUnifyOutputs(int)) );
			menu->setItemParameter(lastIndex, screen->outputsUnified() ? 0 : 1);
			menu->setItemChecked(lastIndex, screen->outputsUnified());
		}
	}
	else
#endif
		populateLegacyMenu(menu);
}

void KRandRSystemTray::populateLegacyMenu(KPopupMenu* menu)
{
	int lastIndex = 0;

	menu->insertTitle(SmallIcon("window_fullscreen"), i18n("Screen Size"));

	LegacyRandRScreen *screen = m_display->currentLegacyScreen();
	Q_ASSERT(screen);

	int numSizes = screen->numSizes();
	int* sizeSort = new int[numSizes];

	for (int i = 0; i < numSizes; i++) 
		sizeSort[i] = screen->pixelCount(i);

	for (int j = 0; j < numSizes; j++) {
		int highest = -1, highestIndex = -1;

		for (int i = 0; i < numSizes; i++) {
			if (sizeSort[i] && sizeSort[i] > highest) {
				highest = sizeSort[i];
				highestIndex = i;
			}
		}
		sizeSort[highestIndex] = -1;
		Q_ASSERT(highestIndex != -1);

		lastIndex = menu->insertItem(i18n("%1 x %2").arg(screen->pixelSize(highestIndex).width()).arg(screen->pixelSize(highestIndex).height()));

		if (screen->proposedSize() == highestIndex)
			menu->setItemChecked(lastIndex, true);

		menu->setItemParameter(lastIndex, highestIndex);
		menu->connectItem(lastIndex, this, SLOT(slotResolutionChanged(int)));
	}
	delete [] sizeSort;
	sizeSort = 0L;

	// Don't display the rotation options if there is no point (ie. none are supported)
	// XFree86 4.3 does not include rotation support.
	if (screen->rotations() != RandR::Rotate0) {
		menu->insertTitle(SmallIcon("reload"), i18n("Orientation"));

		for (int i = 0; i < 6; i++) {
			if ((1 << i) & screen->rotations()) {
				lastIndex = menu->insertItem(RandR::rotationIcon(1 << i, screen->rotation()), RandR::rotationName(1 << i));

				if (m_display->currentLegacyScreen()->proposedRotation() & (1 << i))
					menu->setItemChecked(lastIndex, true);

				menu->setItemParameter(lastIndex, 1 << i);
				menu->connectItem(lastIndex, this, SLOT(slotOrientationChanged(int)));
			}
		}
	}

	QStringList rr = screen->refreshRates(screen->proposedSize());

	if (rr.count())
		menu->insertTitle(SmallIcon("clock"), i18n("Refresh Rate"));

	int i = 0;
	for (QStringList::Iterator it = rr.begin(); it != rr.end(); ++it, i++) {
		lastIndex = menu->insertItem(*it);

		if (screen->proposedRefreshRate() == i)
			menu->setItemChecked(lastIndex, true);

		menu->setItemParameter(lastIndex, i);
		menu->connectItem(lastIndex, this, SLOT(slotRefreshRateChanged(int)));
	}
}

void KRandRSystemTray::populateRotations(KPopupMenu *menu, int rotations, int rotation, QObject *receiver, const char *member)
{
	int lastIndex;
	for (int i = 0; i < 6; i++) 
	{
		if ((1 << i) & rotations) 
		{
			lastIndex = menu->insertItem(RandR::rotationIcon(1 << i, rotation),
							   RandR::rotationName(1 << i));

			menu->setItemParameter(lastIndex, 1 << i);
			menu->connectItem(lastIndex, receiver, member);
			if (rotation & (1 << i))
				menu->setItemChecked(lastIndex, true);
		}
	}
}

void KRandRSystemTray::populateSizes(KPopupMenu *menu, SizeList sizes, QSize size, QObject *receiver, const char *member)
{
	int lastIndex;
	for (int i = 0; i < sizes.count(); ++i) 
	{
		QSize s = sizes[i];
		lastIndex = menu->insertItem(QString("%1 x %2").arg(s.width()).arg(s.height()));
		menu->setItemParameter(lastIndex, i);
		menu->connectItem(lastIndex, receiver, member);

		if (s == size) 
			menu->setItemChecked(lastIndex, true);
	}
}

void KRandRSystemTray::populateRates(KPopupMenu *menu, RateList rates, float rate, QObject *receiver, const char *member)
{
	int lastIndex, i = 0;
	RateList::const_iterator it;
	for (it = rates.begin(); it != rates.end(); ++it)
	{
		lastIndex = menu->insertItem(i18n("%1 Hz").arg(QString::number(*it, 'f', 1)));
		menu->setItemParameter(lastIndex, i);
		if (*it == rate)
			menu->setItemChecked(lastIndex, true);

		menu->connectItem(lastIndex, receiver, member);
		++i;
	}
}

void KRandRSystemTray::slotResolutionChanged(int parameter)
{
	int index = parameter;
	LegacyRandRScreen *screen = m_display->currentLegacyScreen();
	Q_ASSERT(screen);

	if (screen->size() == index)
		return;

	screen->proposeSize(index);

	screen->proposeRefreshRate(-1);

	if (screen->applyProposedAndConfirm()) 
	{
		KConfig config("kcmrandrrc");
		if (m_display->syncTrayApp(config))
			screen->save(config);
	}
}

void KRandRSystemTray::slotOrientationChanged(int parameter)
{
	LegacyRandRScreen *screen = m_display->currentLegacyScreen();
	Q_ASSERT(screen);
	
	int propose = screen->rotation();
	int rotate = parameter;

	if (rotate & RandR::RotateMask)
		propose &= RandR::ReflectMask;

	propose ^= rotate;

	if (screen->rotation() == propose)
		return;

	screen->proposeRotation(propose);

	if (screen->applyProposedAndConfirm()) 
	{
		KConfig config("kcmrandrrc");
		if (m_display->syncTrayApp(config))
			screen->save(config);
	}
}

void KRandRSystemTray::slotRefreshRateChanged(int parameter)
{
	LegacyRandRScreen *screen = m_display->currentLegacyScreen();
	Q_ASSERT(screen);
	
	int index = parameter;

	if (screen->refreshRate() == index)
		return;

	screen->proposeRefreshRate(index);

	if (screen->applyProposedAndConfirm())
	{
		KConfig config("kcmrandrrc");
		if (m_display->syncTrayApp(config))
			screen->save(config);
	}
}

void KRandRSystemTray::slotPrefs()
{
	KCMultiDialog *kcm = new KCMultiDialog( KDialogBase::Plain, i18n( "Configure" ), this );

	kcm->addModule( "display" );
	kcm->setPlainCaption( i18n( "Configure Display" ) );
	kcm->exec();
}

void KRandRSystemTray::slotToggleRunStartup()
{
	m_runStartup = !m_runStartup;
	KConfig cfg("kcmrandrrc");

	cfg.setGroup("General");
	cfg.writeEntry("AutoStart", m_runStartup);
}

void KRandRSystemTray::slotSwitchDisplay()
{
    m_display->switchOutput();
}
