/*
 * Copyright (c) 2007      Gustavo Pichorim Boiko <gustavo.boiko@kdemail.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <kdebug.h>
#include <kconfig.h>
#include "randrscreen.h"
#include "randrcrtc.h"
#include "randroutput.h"
#include "randrmode.h"

#ifdef HAS_RANDR_1_2
RandRScreen::RandRScreen(int screenIndex)
	: m_resources(0L)
{
	m_index = screenIndex;
	m_rect = QRect(0, 0, XDisplayWidth(qt_xdisplay(), m_index), XDisplayHeight(qt_xdisplay(), m_index));

	m_connectedCount = 0;
	m_activeCount = 0;

	loadSettings();
	load();

	// select for randr input events
	int mask = RRScreenChangeNotifyMask | 
		   RRCrtcChangeNotifyMask | 
		   RROutputChangeNotifyMask | 
		   RROutputPropertyNotifyMask;
	XRRSelectInput(qt_xdisplay(), rootWindow(), 0);
	XRRSelectInput(qt_xdisplay(), rootWindow(), mask); 
}

RandRScreen::~RandRScreen()
{
	if (m_resources)
		XRRFreeScreenResources(m_resources);
}

int RandRScreen::index() const
{
	return m_index;
}

XRRScreenResources* RandRScreen::resources() const
{
	return m_resources;
}

Window RandRScreen::rootWindow() const
{
	return RootWindow(qt_xdisplay(), m_index);
}

void RandRScreen::loadSettings(bool notify)
{
	bool changed = false;
	int minW, minH, maxW, maxH;

	Status status = XRRGetScreenSizeRange(qt_xdisplay(), rootWindow(),
					 &minW, &minH, &maxW, &maxH);
	//FIXME: we should check the status here
	Q_UNUSED(status);
	QSize minSize = QSize(minW, minH);
	QSize maxSize = QSize(maxW, maxH);

	if (minSize != m_minSize || maxSize != m_maxSize)
	{
		m_minSize = minSize;
		m_maxSize = maxSize;
		changed = true;
	}

	if (m_resources)
		XRRFreeScreenResources(m_resources);

	m_resources = XRRGetScreenResources(qt_xdisplay(), rootWindow());
	Q_ASSERT(m_resources);

	if (RandR::timestamp != m_resources->timestamp)
		RandR::timestamp = m_resources->timestamp;

	// get all modes
	for (int i = 0; i < m_resources->nmode; ++i)
	{
		if (!m_modes.contains(m_resources->modes[i].id))
		{
			m_modes[m_resources->modes[i].id] = RandRMode(&m_resources->modes[i]);
			changed = true;
		}
	}

	//get all crtcs
	for (int i = 0; i < m_resources->ncrtc; ++i)
	{
		if (m_crtcs.contains(m_resources->crtcs[i]))
			m_crtcs[m_resources->crtcs[i]]->loadSettings(notify);
		else
		{
			RandRCrtc *c = new RandRCrtc(this, m_resources->crtcs[i]);
			connect(c, SIGNAL(crtcChanged(RRCrtc, int)), this, SIGNAL(configChanged()));
			m_crtcs[m_resources->crtcs[i]] = c;
			changed = true;
		}
	}

	//get all outputs
	for (int i = 0; i < m_resources->noutput; ++i)
	{
		if (m_outputs.contains(m_resources->outputs[i]))
			m_outputs[m_resources->outputs[i]]->loadSettings(true);
		else
		{
			RandROutput *o = new RandROutput(this, m_resources->outputs[i]);
			connect(o, SIGNAL(outputChanged(RROutput, int)), this, SLOT(slotOutputChanged(RROutput, int)));
			m_outputs[m_resources->outputs[i]] = o;
			if (o->isConnected())
				m_connectedCount++;
			if (o->isActive())
				m_activeCount++;

			changed = true;
		}
	}

	if (notify && changed)
		emit configChanged();
}

void RandRScreen::handleEvent(XRRScreenChangeNotifyEvent* event)
{
	m_rect.setWidth(event->width);
	m_rect.setHeight(event->height);

	emit configChanged();
}

void RandRScreen::handleRandREvent(XRRNotifyEvent* event)
{
	RandRCrtc *c;
	RandROutput *o;
	XRRCrtcChangeNotifyEvent *crtcEvent;
	XRROutputChangeNotifyEvent *outputEvent;
	XRROutputPropertyNotifyEvent *propertyEvent;

	// forward events to crtcs and outputs
	switch (event->subtype) {
		case RRNotify_CrtcChange:
			crtcEvent = (XRRCrtcChangeNotifyEvent*)event;
			c = crtc(crtcEvent->crtc);
			Q_ASSERT(c);
			c->handleEvent(crtcEvent);
			return;

		case RRNotify_OutputChange:
			outputEvent = (XRROutputChangeNotifyEvent*)event;
			o = output(outputEvent->output);
			Q_ASSERT(o);
			o->handleEvent(outputEvent);
			return;

		case RRNotify_OutputProperty:
			propertyEvent = (XRROutputPropertyNotifyEvent*)event;
			o = output(propertyEvent->output);
			Q_ASSERT(o);
			o->handlePropertyEvent(propertyEvent);
			return;
	}	
}

QSize RandRScreen::minSize() const
{
	return m_minSize;
}

QSize RandRScreen::maxSize() const
{
	return m_maxSize;
}

CrtcMap RandRScreen::crtcs() const
{
	return m_crtcs;
}

RandRCrtc* RandRScreen::crtc(RRCrtc id) const
{
	if (m_crtcs.contains(id))
		return m_crtcs[id];

	return NULL;
}

OutputMap RandRScreen::outputs() const
{
	return m_outputs;
}

RandROutput* RandRScreen::output(RROutput id) const
{
	if (m_outputs.contains(id))
		return m_outputs[id];

	return NULL;
}

ModeMap RandRScreen::modes() const
{
	return m_modes;
}

RandRMode RandRScreen::mode(RRMode id) const
{
	if (m_modes.contains(id))
		return m_modes[id];

	return RandRMode(0);
}

bool RandRScreen::adjustSize(QRect minimumSize)
{
	//try to find a size in which all outputs fit
	
	//start with a given minimum rect
	QRect rect = QRect(0,0,0,0).unite(minimumSize);

	OutputMap::const_iterator it;
	for (it = m_outputs.constBegin(); it != m_outputs.constEnd(); ++it)
	{
		// outputs that are not active should not be taken into account
		// when calculating the screen size
		RandROutput *o = (*it);
		if (!o->isActive())
			continue;
		rect = rect.unite(o->rect());
	}


	// check bounds
	if (rect.width() < m_minSize.width())
		rect.setWidth(m_minSize.width());
	if (rect.height() < m_minSize.height())
		rect.setHeight(m_minSize.height());

	if (rect.width() > m_maxSize.width())
		return false;
	if (rect.height() > m_maxSize.height())
		return false;

	return setSize(rect.size());
}

bool RandRScreen::setSize(QSize s)
{
	if (s == m_rect.size())
		return true;

	if (s.width() < m_minSize.width() || 
	    s.height() < m_minSize.height() ||
	    s.width() > m_maxSize.width() ||
	    s.height() > m_maxSize.height())
		return false;

	int widthMM, heightMM;
	float dpi;

	/* values taken from xrandr */
	dpi = (25.4 * DisplayHeight(qt_xdisplay(), m_index)) / DisplayHeightMM(qt_xdisplay(), m_index);
	widthMM =  (int) ((25.4 * s.width()) / dpi);
	heightMM = (int) ((25.4 * s.height()) / dpi);

	XRRSetScreenSize(qt_xdisplay(), rootWindow(), s.width(), s.height(), widthMM, heightMM);
	m_rect.setSize(s);

	return true;
}

int RandRScreen::connectedCount() const
{
	return m_connectedCount;
}

int RandRScreen::activeCount() const
{
	return m_activeCount;
}

bool RandRScreen::outputsUnified() const
{
	return m_outputsUnified;
}

int RandRScreen::unifiedRotations() const
{
	bool first = true;
	int rotations = RandR::Rotate0;

	CrtcMap::const_iterator it;
	for (it = m_crtcs.constBegin(); it != m_crtcs.constEnd(); ++it)
	{
		if (!(*it)->connectedOutputs().count())
			continue;

		if (first)
		{
			rotations = (*it)->rotations();
			first = false;
		}
		else
			rotations &= (*it)->rotations();
	}

	return rotations;
}

SizeList RandRScreen::unifiedSizes() const
{
	SizeList sizeList;
	bool first = true;
	OutputMap::const_iterator it;

	for (it = m_outputs.constBegin(); it != m_outputs.constEnd(); ++it)
	{
		if (!(*it)->isActive())
			continue;

		if (first)
		{
			// we start using the list from the first output
			sizeList = (*it)->sizes();
			first = false;
		}
		else
		{
			SizeList outputSizes = (*it)->sizes();
			SizeList::iterator s;
			for (int i = sizeList.count() - 1; i >=0; --i)
			{
				// check if the current output has the i-th size of the sizeList
				// if not, remove from the list
				if (outputSizes.find(sizeList[i]) == outputSizes.end())
					sizeList.remove(sizeList[i]);
			}
		}
	}

	return sizeList;
}

QRect RandRScreen::rect() const
{
	return m_rect;
}

void RandRScreen::load(KConfig &config)
{
	config.setGroup("Screen_" + QString::number(m_index));
	m_outputsUnified = config.readBoolEntry("OutputsUnified", true);
	m_unifiedRect = config.readRectEntry("UnifiedRect", &m_unifiedRect);
	m_unifiedRotation = config.readNumEntry("UnifiedRotation", (int) RandR::Rotate0);

	slotUnifyOutputs(m_outputsUnified);

	OutputMap::iterator it;
	for (it = m_outputs.begin(); it != m_outputs.end(); ++it)
	{
		if ((*it)->isConnected())
			(*it)->load(config);
	}
}

void RandRScreen::save(KConfig &config)
{
	config.setGroup("Screen_" + QString::number(m_index));
	config.writeEntry("OutputsUnified", m_outputsUnified);
	config.writeEntry("UnifiedRect", m_unifiedRect);
	config.writeEntry("UnifiedRotation", m_unifiedRotation);
	OutputMap::iterator it;
	for (it = m_outputs.begin(); it != m_outputs.end(); ++it)
	{
		if ((*it)->isConnected())
			(*it)->save(config);
	}
}

void RandRScreen::save()
{
	KConfig cfg("kcmrandrrc");
	save(cfg);
}

void RandRScreen::load()
{
	KConfig cfg("kcmrandrrc");
	load(cfg);
}

bool RandRScreen::applyProposed(bool confirm)
{
	bool succeed = true;
	QRect r;
	OutputMap::iterator it;
	for (it = m_outputs.begin(); it != m_outputs.end(); ++it)
	{
		RandROutput *output = *it;
		r = output->rect();
		if (!output->applyProposed())
		{
			succeed = false;
			break;
		}
	}

	// if we could apply the config clean, ask for confirmation
	if (succeed && confirm)
		succeed = RandR::confirm(r);

	// if we succeded applying and the user confirmer the changes,
	// just return from here
	if (succeed)
		return true;

	//Revert changes if not succeed
	for (it = m_outputs.begin(); it != m_outputs.end(); ++it)
	{
		RandROutput *o = *it;
		if (o->isConnected())
		{
			o->proposeOriginal();
			o->applyProposed();
		}
	}
	return false;
}

void RandRScreen::unifyOutputs()
{
	SizeList sizes = unifiedSizes();

	//FIXME: better handle this
	if (!sizes.count())
		return;

	if (sizes.findIndex(m_unifiedRect.size()) == -1)
		m_unifiedRect.setSize(sizes.first());

	// iterate over all outputs and make sure all connected outputs get activated
	// and use the right size
	OutputMap::iterator it;
	for(it = m_outputs.begin(); it != m_outputs.end(); ++it)
	{
		RandROutput *o = *it;
		// if the output is not connected we don't need to do anything
		if (!o->isConnected())
			continue;

		// if the output is connected and already has the same rect and rotation
		// as the unified ones, continue
		if (o->isActive() && o->rect() == m_unifiedRect
						  && o->rotation() == m_unifiedRotation)
			continue;

		o->proposeRect(m_unifiedRect);
		o->proposeRotation(m_unifiedRotation);
		o->applyProposed(RandR::ChangeRect |
						 RandR::ChangeRotation, false);
	}

	// FIXME: if by any reason we were not able to unify the outputs, we should
	// do something

	save();
}

void RandRScreen::slotResizeUnified(int index)
{
	m_unifiedRect.setSize(unifiedSizes()[index]);
	unifyOutputs();
}

void RandRScreen::slotUnifyOutputs(int unified)
{
	m_outputsUnified = unified;
	KConfig cfg("kcmrandrrc");

	if (!unified || m_connectedCount <= 1)
	{
		OutputMap::iterator it;
		for (it = m_outputs.begin(); it != m_outputs.end(); ++it)
		{
			RandROutput *output = *it;
			if (output->isConnected())
			{
				output->load(cfg);
				output->applyProposed();
			}
		}
	}
	else
	{
		SizeList sizes = unifiedSizes();

		if (!sizes.count())
		{
			// FIXME: this should be better handle
			return;
		}

		QSize s = m_unifiedRect.size();

		// if the last size we used is not available, use the first one
		// from the list
		if (sizes.findIndex(s) == -1)
			s = sizes[0];

		m_unifiedRect.setTopLeft(QPoint(0,0));
		m_unifiedRect.setSize(s);
		unifyOutputs();
	}
}

void RandRScreen::slotRotateUnified(int index)
{
	m_unifiedRotation = index;

	unifyOutputs();
}

void RandRScreen::slotOutputChanged(RROutput id, int changes)
{
	kdDebug() << "[SCREEN] Got output change " << changes << endl;

	int connected = 0, active = 0;
	OutputMap::iterator it;
	for(it = m_outputs.begin(); it != m_outputs.end(); ++it)
	{
		RandROutput *output = *it;
		if (output->isConnected())
			connected++;
		if (output->isActive())
			active++;
	}

	m_connectedCount = connected;
	m_activeCount = active;

	// if there is less than 2 outputs connected, there is no need to unify
	if (connected <= 1)
		return;

	RandROutput *o = m_outputs[id];
	Q_ASSERT(o);

	// if we are using a unified layout, we should not allow an output
	// to use another position or size. So ask it to come back to the unified
	// size.
	if (m_outputsUnified)
	{
		if (o->rect() != m_unifiedRect || o->rotation() != m_unifiedRotation)
			unifyOutputs();
	}

	// TODO: handle the changes not to allow overlapping on non-unified
	// setups
	
	save();
	emit configChanged();
}

void RandRScreen::switchOutput()
{
	//refresh settings so that the new monitor gets detected
	loadSettings();

	// if there is just one connected output, do not switch
	if (connectedCount() < 2)
		return;

	// disable all outputs and activate the first connected one
	if (activeCount() > 1)
	{
		bool first = true;
		slotUnifyOutputs(0);
		OutputMap::iterator it;
		for(it = m_outputs.begin(); it != m_outputs.end(); ++it)
		{
			RandROutput *output = *it;
			if (first && output->isActive())
				first = false;
			else if (first && output->isConnected())
			{
				SizeList sizes = output->sizes();
				if (!sizes.count())
					continue;

				output->proposeRect(QRect(QPoint(0,0), sizes.first()));
				if (output->applyProposed(false))
					first = false;
				else
					output->slotDisable();
			}
			else
				output->slotDisable();
		}

		for(it = m_outputs.begin(); it != m_outputs.end(); ++it)
		{
			RandROutput *output = *it;
			if (output->isConnected())
			{
				SizeList sizes = output->sizes();
				if (!sizes.count())
					continue;

				output->proposeRect(QRect(QPoint(0,0), sizes.first()));
				if (output->applyProposed(RandR::ChangeRect, false))
					break;
			}
		}
	}
	else
	{
		OutputMap::iterator it;
		// first disable the currently active output and try to active the next
		// connected one
		bool found_active = false, found_connected = false;
		for(it = m_outputs.begin(); it != m_outputs.end(); ++it)
		{
			RandROutput *output = *it;
			if (!found_active && output->isActive())
			{
				output->slotDisable();
				found_active = true;
			}
			else if (found_active && output->isConnected())
			{
				SizeList sizes = output->sizes();
				if (!sizes.count())
					continue;

				output->proposeRect(QRect(QPoint(0,0), sizes.first()));
				if (output->applyProposed(RandR::ChangeRect, false))
				{
					found_connected = true;
					break;
				}
			}
		}

		// if we could not find the next connected output after the active one,
		// activate all
		if (!found_connected)
			slotUnifyOutputs(1);
	} // if activeCount < 2
}

#include "randrscreen.moc"

#endif

